class Logger:

    def __init__(self):
        self.dlog = None

    def start(self, dlog, config=None, delete_existing=False):
        self.dlog = dlog

    def log(self, content: dict):
        raise NotImplementedError()

    def load_logs(self, ignore=tuple()):
        raise NotImplementedError()

    def finish(self):
        pass



from .plotter import Plotter
from matplotlib import pyplot as plt
import pandas as pd
import seaborn as sns


class LinePlotter(Plotter):

    def plot_group(self, data, x, y, label, ax, xpid=None, read_every=1, smooth_window=10, align_x=1, alpha=0.4, linewidth=3):
        color = next(ax._get_lines.prop_cycler)['color']
        if xpid is not None:
            all_data = []
            unique_xpids = sorted(data[xpid].unique())
            for xid in unique_xpids:
                xp_data = data[data[xpid] == xid][::read_every]
                xp_data[y+'_smooth'] = xp_data[y].rolling(smooth_window, min_periods=smooth_window).mean()
                all_data.append(xp_data)
                ax.plot(xp_data[x], xp_data[y+'_smooth'], linestyle='dashed', color=color, label='_nolegend_', alpha=alpha)
            data = pd.concat(all_data)
        else:
            data = data[::read_every]
            data[y+'_smooth'] = data[y].rolling(smooth_window, min_periods=smooth_window).mean()
        sns.lineplot(data=data, x=x, y=y+'_smooth', label=label, ax=ax, color=color, linewidth=linewidth)

    def plot(self, data, x, y, group=None, xpid=None, read_every=1, smooth_window=10, align_x=1, ax=None, linewidth=3, alpha=0.4):
        data = pd.DataFrame(data)
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 10))

        if group is None:
            self.plot_group(data=data, x=x, y=y, ax=ax, xpid=xpid, read_every=read_every, smooth_window=smooth_window, align_x=align_x, alpha=alpha, linewidth=linewidth)
        else:
            unique_groups = sorted(data[group].unique())
            for g in unique_groups:
                group_data = data[data[group] == g]
                self.plot_group(data=group_data, x=x, y=y, label=g, ax=ax, xpid=xpid, read_every=read_every, smooth_window=smooth_window, align_x=align_x)
        return ax



from .logger import Logger
import wandb


class WandbLogger(Logger):

    def __init__(self, name):
        super().__init__()
        self.name = name
        self.wandb_config = None

    def start(self, dlog, config=None, delete_existing=False):
        super().start(dlog, config=config, delete_existing=delete_existing)
        wandb.init(project=self.name, config=config)
        self.wandb_config = wandb.config

    def log(self, content: dict):
        wandb.log(content)

    def finish(self):
        wandb.finish()
        self.wandb_config = None



import os
import glob
import datetime
import logging
import ujson as json


class Experiment:

    def __init__(self, config, loggers=tuple(), name_field='project', logdir_field='logdir', step=0):
        self.name_field = name_field
        self.logdir_field = logdir_field
        self.config = config
        self.loggers = loggers
        self.step = step
        self.started = False

    @property
    def name(self):
        return self.config[self.name_field]

    @property
    def logdir(self):
        return self.config[self.logdir_field]

    @property
    def expdir(self):
        return os.path.join(self.logdir, self.name)

    @property
    def explog(self):
        return os.path.join(self.expdir, 'exp.json')

    @classmethod
    def from_namespace(cls, args, name_field='project', logdir_field='logdir', loggers=tuple()):
        return cls(vars(args), name_field=name_field, logdir_field=logdir_field, loggers=loggers)

    @classmethod
    def from_fconfig(cls, fname):
        with open(fname) as f:
            d = json.load(f)
            del d['time']
            return cls(**d)

    def save(self):
        with open(self.explog, 'wt') as f:
            json.dump(dict(
                name_field=self.name_field,
                logdir_field=self.logdir_field,
                config=self.config,
                step=self.step,
                time=datetime.datetime.utcnow().isoformat(),
            ), f, indent=2)

    def load(self):
        with open(self.explog) as f:
            d = json.load(f)
            for k, v in d.items():
                setattr(self, k, v)

    def start(self, delete_existing=False):
        if self.exists():
            if not delete_existing and os.path.isfile(self.explog):
                logging.critical('Resuming from {}'.format(self.explog))
                self.load()
        else:
            logging.critical('Making directory at {}'.format(self.expdir))
            os.makedirs(self.expdir)
        for logger in self.loggers:
            logger.start(self.expdir, self.config, delete_existing=delete_existing)
        self.started = True
        return self

    def load_logs(self, logger, ignore=('time',)):
        assert self.exists(), 'Experiment does not exist at {}'.format(self.expdir)
        logger.start(self.expdir, self.config, delete_existing=False)
        logs = logger.load_logs(ignore=ignore)
        for d in logs:
            d.update(self.config)
        logger.finish()
        return logs

    @classmethod
    def discover_logs(self, glob_path, logger, ignore=('time',)):
        dirs = glob.glob(glob_path)
        data = []
        for d in dirs:
            f = os.path.join(d, 'exp.json')
            if os.path.isfile(f):
                exp = Experiment.from_fconfig(f)
                data.extend(exp.load_logs(logger, ignore=ignore))
        return data

    def exists(self):
        return os.path.isdir(self.expdir)

    def log(self, content: dict):
        assert self.started, 'Please run experiment.start()'
        content['step'] = self.step
        content['time'] = datetime.datetime.utcnow().isoformat()
        for logger in self.loggers:
            logger.log(content)
        self.save()
        self.step += 1

    def finish(self):
        for logger in self.loggers:
            logger.finish()



from .logger import Logger
import os
import ujson as json


class JSONLogger(Logger):

    def __init__(self, logname='log.jsonl'):
        super().__init__()
        self.logname = logname
        self.fname = None

    def start(self, dlog, config=None, delete_existing=False):
        super().start(dlog, config=config, delete_existing=delete_existing)
        self.fname = os.path.join(self.dlog, self.logname)
        if delete_existing and os.path.isfile(self.fname):
            os.remove(self.fname)

    def log(self, content: dict):
        with open(self.fname, 'at') as f:
            f.write(json.dumps(content) + '\n')

    def load_logs(self, ignore=tuple()):
        logs = []
        with open(self.fname, 'rt') as f:
            for line in f:
                d = {k: v for k, v in json.loads(line).items() if k not in ignore}
                logs.append(d)
        return logs

    def finish(self):
        self.name = None
